package ac.essex.ooechs.fractals.maths;

import ac.essex.ooechs.fractals.colouring.Colourer;

import java.awt.image.BufferedImage;
import java.awt.*;

/**
 * Converts a given fractal function into an image
 *
 * @author Olly Oechsle, University of Essex, Date: 22-Nov-2007
 * @version 1.0
 */
public abstract class Fractal {

    public double xmin = -1.6140;
    public double xmax = 0.5196;
    public double ymin = -1.3207;
    public double ymax = 1.3319;

    public int max_iterations = 128;

    public boolean hires = false;

    public void render(BufferedImage image, Colourer c) {

        double xRange = xmax - xmin;
        double yRange = ymax - ymin;

        double imageWidth = image.getWidth();
        double imageHeight = image.getHeight();

        // X has to be converted from screen coordinates
        // cache these values to avoid recomputation
        double[] xCache = new double[image.getWidth()];
        for (int imageX = 0; imageX < imageWidth; imageX++) {
            xCache[imageX] = xmin + ((imageX / imageWidth) * xRange);
        }

        // for hires
        double halfX = (0.5 / imageWidth) * xRange;
        double halfY = (0.5 / imageHeight) * yRange;

        for (int imageY = 0; imageY < imageHeight; imageY++) {
            double y = ymin + ((imageY / imageHeight) * yRange);
            for (int imageX = 0; imageX < imageWidth; imageX++) {

                double iterations = getIterations(xCache[imageX], y, max_iterations);

                if (hires) {
                    iterations += getIterations(xCache[imageX] + halfX, y, max_iterations);
                    iterations += getIterations(xCache[imageX] - halfX, y, max_iterations);
                    iterations += getIterations(xCache[imageX], y + halfY, max_iterations);
                    iterations += getIterations(xCache[imageX], y - halfY, max_iterations);
                    iterations /= 5;
                }

                image.setRGB(imageX, imageY, c.getColor(iterations, max_iterations).getRGB());
                
            }
        }

    }

    public abstract double getIterations(double x, double y, final int max_iterations);

    public void center(int imageX, int imageY, BufferedImage image) {
        zoom(imageX, imageY, image, 1.0);
    }

    public void zoom(int imageX, int imageY, BufferedImage image, double zoom) {

        double xRange = xmax - xmin;
        double yRange = ymax - ymin;

        double imageWidth = image.getWidth();
        double imageHeight = image.getHeight();

        double x = xmin + ((imageX / imageWidth) * xRange);
        double y = ymin + ((imageY / imageHeight) * yRange);

        double halfXrange = (xRange * zoom) / 2;
        double halfYrange = (yRange * zoom) / 2;

        xmin = x - halfXrange;
        xmax = x + halfXrange;
        ymin = y - halfYrange;
        ymax = y + halfYrange;

    }

    public void zoom(Rectangle r, BufferedImage image) {

        double imageWidth = image.getWidth();
        double imageHeight = image.getHeight();

        double xRange = xmax - xmin;
        double yRange = ymax - ymin;

        double newXRange = (r.width / imageWidth) * xRange;
        double newYRange = (r.height / imageHeight) * yRange;

        xmin = xmin + ((r.x / imageWidth) * xRange);
        ymin = ymin + ((r.y / imageHeight) * yRange);

        xmax = xmin + newXRange;
        ymax = ymin + newYRange;

    }


}
